@@ -20,6 +20,8 @@
 
  *****************************************************************************)
 
+module Pcre = Re.Pcre
+
 type fd = Unix.file_descr
 
 external poll :
@@ -64,6 +66,7 @@ type 'a t = {
 }
 
 type 'a scheduler = {
+  on_error : exn -> Printexc.raw_backtrace -> unit;
   out_pipe : Unix.file_descr;
   in_pipe : Unix.file_descr;
   compare : 'a -> 'a -> int;
@@ -84,9 +87,10 @@ let clear_tasks s =
   s.tasks <- [];
   Mutex.unlock s.tasks_m
 
-let create ?(compare = compare) () =
+let create ?(on_error = Printexc.raise_with_backtrace) ?(compare = compare) () =
   let out_pipe, in_pipe = Unix.pipe () in
   {
+    on_error;
     out_pipe;
     in_pipe;
     compare;
@@ -126,7 +130,8 @@ module Task = struct
       enrich =
         (fun e ->
           List.fold_left
-            (fun e -> function `Delay s -> { e with t = min e.t (t0 +. s) }
+            (fun e -> function
+              | `Delay s -> { e with t = min e.t (t0 +. s) }
               | `Read s -> { e with r = s :: e.r }
               | `Write s -> { e with w = s :: e.w }
               | `Exception s -> { e with x = s :: e.x })
@@ -268,13 +273,21 @@ let exec s (priorities : 'a -> bool) =
    * win32 because a thread can double-lock
    * the same mutex.. *)
   if Sys.os_type <> "Win32" then assert (not (Mutex.try_lock s.ready_m));
-  try
-    let (_, task), remaining = remove (fun (p, _) -> priorities p) s.ready in
-    s.ready <- remaining;
-    Mutex.unlock s.ready_m;
-    add_t s (task ());
-    true
-  with Not_found -> false
+  match remove (fun (p, _) -> priorities p) s.ready with
+    | (_, task), remaining ->
+        s.ready <- remaining;
+        Mutex.unlock s.ready_m;
+        let tasks =
+          match task () with
+            | exception exn ->
+                let bt = Printexc.get_raw_backtrace () in
+                s.on_error exn bt;
+                []
+            | v -> v
+        in
+        add_t s tasks;
+        true
+    | exception Not_found -> false
 
 exception Queue_stopped
 exception Queue_processed
@@ -325,7 +338,7 @@ let queue ?log ?(priorities = fun _ -> true) s name =
       Mutex.unlock s.ready_m;
       process s log;
       Mutex.unlock s.select_m;
-      wake ();
+      wake ()
     end
     else begin
       (* We use s.ready_m mutex here.
@@ -354,11 +367,12 @@ let queue ?log ?(priorities = fun _ -> true) s name =
     Condition.signal s.queue_stopped_c;
     Mutex.unlock s.queues_m
   in
-  ( try f () with
+  (try f () with
     | Queue_stopped -> ()
     | exn ->
-        on_done ();
-        raise exn );
+        let bt = Printexc.get_raw_backtrace () in
+        (try on_done () with _ -> ());
+        Printexc.raise_with_backtrace exn bt);
   on_done ()
 
 module Async = struct
@@ -404,8 +418,8 @@ module Async = struct
     try
       begin
         match t.fd with
-        | Some t -> ignore (Unix.write t (Bytes.of_string " ") 0 1)
-        | None -> raise Stopped
+          | Some t -> ignore (Unix.write t (Bytes.of_string " ") 0 1)
+          | None -> raise Stopped
       end;
       Mutex.unlock t.m
     with e ->
@@ -417,10 +431,10 @@ module Async = struct
     try
       begin
         match t.fd with
-        | Some c ->
-            t.stop := true;
-            ignore (Unix.write c (Bytes.of_string " ") 0 1)
-        | None -> raise Stopped
+          | Some c ->
+              t.stop := true;
+              ignore (Unix.write c (Bytes.of_string " ") 0 1)
+          | None -> raise Stopped
       end;
       t.fd <- None;
       Mutex.unlock t.m
@@ -593,7 +607,7 @@ struct
                   [{ priority; events; handler = f }]
               | _ ->
                   exec x;
-                  [] )
+                  [])
         | None -> [{ priority; events; handler = f }]
     in
     (* Catch all exceptions.. *)
@@ -658,12 +672,12 @@ struct
         let n = write pos len in
         if n <= 0 then (
           on_error Io_error;
-          [] )
+          [])
         else if n < len then
           [{ priority; events = [`Write unix_socket]; handler = f (pos + n) }]
         else (
           exec ();
-          [] )
+          [])
       with
         | Unix.Unix_error (Unix.EWOULDBLOCK, _, _) when Sys.os_type = "Win32" ->
             [{ priority; events = [`Write unix_socket]; handler = f pos }]
@@ -818,12 +832,12 @@ module Monad = struct
         if not m.locked then (
           (* I don't think shuffling tasks
            * matters here.. *)
-            match m.tasks with
+          match m.tasks with
             | x :: l ->
                 m.tasks <- l;
                 m.locked <- true;
                 task x :: tasks
-            | _ -> tasks )
+            | _ -> tasks)
         else tasks
 
       let rec handler _ =
